import argparse
import datetime
import os
import re

from Dataloader import Data
import json
import tensorflow as tf
import tensorflow_addons as tfa
import tensorflow_hub as hub
import tensorflow_text as tft
from official.nlp import optimization
from tensorflow.keras.layers import Dense, Embedding, Conv1D, GlobalAveragePooling1D

class Baseline:
    def __init__(self, data, args):
        inputs = tf.keras.layers.Input(shape=(args.sequence_length,))
        if args.pretraining:
            inputs = tf.keras.layers.Input(shape=(), dtype=tf.string, name='text')
            preprocessing_layer = hub.KerasLayer('https://tfhub.dev/tensorflow/bert_en_uncased_preprocess/3', name='preprocessing')
            encoder_inputs = preprocessing_layer(inputs)
            encoder = hub.KerasLayer('https://tfhub.dev/tensorflow/bert_en_uncased_L-12_H-768_A-12/3', trainable=True, name='BERT_encoder')
            pretrained_outputs = encoder(encoder_inputs)
            x = pretrained_outputs['sequence_output']
        else:
            x = Embedding(args.vocab_size+1, args.emb_dim, mask_zero=True)(inputs)
        for _ in range(args.num_layers):
            x = Conv1D(args.layer_dim, args.kernel_size, padding="same", activation="gelu", strides=1)(x)
            x = tf.keras.layers.Dropout(args.dropout)(x)
        x = GlobalAveragePooling1D()(x)
        outputs = Dense(args.num_labels)(x)

        self._model = tf.keras.Model(inputs=inputs, outputs=outputs)
        optimizer = tfa.optimizers.LazyAdam(args.learning_rate)
        if args.pretraining:
            init_lr = 3e-5
            steps_per_epoch = tf.data.experimental.cardinality(data['train']).numpy()
            num_train_steps = steps_per_epoch * args.epochs
            num_warmup_steps = int(0.1*num_train_steps)
            optimizer = optimization.create_optimizer(
                            init_lr=init_lr,
                            num_train_steps=num_train_steps,
                            num_warmup_steps=num_warmup_steps,
                            optimizer_type='adamw')
        self._model.compile(
                optimizer=optimizer,
                loss=tf.losses.BinaryFocalCrossentropy(from_logits=True, label_smoothing=args.label_smoothing),
                metrics=[tf.metrics.BinaryAccuracy(name="accuracy")])
    
    def train(self, data, args):
        self._model.fit(
            data['train'],
            batch_size=args.batch_size,
            epochs=args.epochs,
            validation_data=(data['val'])
            )

    def test(self, data, args):
        test_logs = self._model.evaluate(data['test'], batch_size=args.batch_size)
        return test_logs[self._model.metrics_names.index("accuracy")]


def main():
    # Parse arguments
    parser = argparse.ArgumentParser()
    parser.add_argument("--dataset", default="imdb", type=str, help="Dataset (imdb, sst2, arxiv).")
    parser.add_argument("--batch_size", default=32, type=int, help="Batch size.")
    parser.add_argument("--epochs", default=3, type=int, help="Number of epochs.")
    parser.add_argument("--learning_rate", default=0.001, type=float, help="Learning rate value.")
    parser.add_argument("--layer_dim", default=64, type=int, help="Layer dimension.")
    parser.add_argument("--num_layers", default=8, type=int, help="Layer count.")
    parser.add_argument("--emb_dim", default=64, type=int, help="Embedding dimension.")
    parser.add_argument("--kernel_size", default=3, type=int, help="Kernel size.")
    parser.add_argument("--dropout", default=0.1, type=float, help="Dropout regularization value.")
    parser.add_argument("--label_smoothing", default=0.3, type=float, help="Label smoothing value.")
    parser.add_argument("--seed", default=42, type=int, help="Random seed.")
    parser.add_argument("--threads", default=1, type=int, help="Maximum number of threads to use.")
    parser.add_argument("--pretraining", default=False, action="store_true", help="Use pretrained embeddings.")
    parser.add_argument("--verbose", default=False, action="store_true", help="Verbose TF logging.")
    args = parser.parse_args([] if "__file__" not in globals() else None)

    # Fix random seeds and threads
    tf.random.set_seed(args.seed)
    tf.config.threading.set_inter_op_parallelism_threads(args.threads)
    tf.config.threading.set_intra_op_parallelism_threads(args.threads)

    # Report only errors by default
    if not args.verbose:
        os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

    # set environment variable for tfhub
    os.environ["TFHUB_CACHE_DIR"] = "~/tmp"

    # Create logdir name
    timestamp = datetime.datetime.now().strftime("%Y-%m-%d_%H%M%S")
    args.logdir = os.path.join("logs", "{}-{}-{}".format(
        os.path.basename(globals().get("__file__", "notebook")),
        timestamp,
        ",".join(("{}={}".format(re.sub("(.)[^_]*_?", r"\1", key), value) for key, value in sorted(vars(args).items())))
    ))

    # Load the data
    vectorize = not args.pretraining
    dataloader = Data(args.dataset, vectorize=vectorize)
    data = dataloader.load()
    args.num_labels = dataloader.num_labels
    args.vocab_size = dataloader.vocab_size
    args.sequence_length = dataloader.sequence_length

    # Create the network and train
    baseline = Baseline(data, args)
    baseline.train(data, args)

    # Compute test set accuracy and print it
    accuracy = baseline.test(data, args)
    print("{:.2f}".format(100 * accuracy))
    parameters = vars(args)
    parameters['accuracy'] = accuracy
    with open(f"baseline_{args.dataset}_results_{timestamp}.json", "w") as out_file:
        json.dump(parameters, out_file, indent=4, sort_keys=True)


if __name__ == "__main__":
    main()
    